
#pragma once

#include "array.hpp"
#include "span.hpp"

namespace ox {

class AnyPtr {
	private:
		struct WrapBase {
			virtual ~WrapBase() = default;
			virtual WrapBase *copyTo(ox::Span<char> s) noexcept = 0;
			virtual operator bool() const noexcept = 0;
		};

		template<typename T>
		struct Wrap: public WrapBase {
			T *data{};
			constexpr Wrap(T *pData) noexcept: data(pData) {
			}
			inline WrapBase *copyTo(ox::Span<char> s) noexcept override {
				return new(s.data()) Wrap{data};
			}
			constexpr operator bool() const noexcept override {
				return data != nullptr;
			}
		};

		WrapBase *m_wrapPtr{};
		ox::Array<char, sizeof(Wrap<void*>)> m_wrapData;
	public:
		constexpr AnyPtr() noexcept = default;
		template<typename T>
		inline AnyPtr(T *ptr) noexcept {
			m_wrapPtr = new(m_wrapData.data()) Wrap(ptr);
		}
		inline AnyPtr(AnyPtr const&other) noexcept {
			if (other) {
				m_wrapPtr = other.m_wrapPtr->copyTo(m_wrapData);
			}
		}
		template<typename T>
		inline AnyPtr &operator=(T *ptr) noexcept {
			m_wrapPtr = new(m_wrapData.data()) Wrap(ptr);
			return *this;
		}
		inline AnyPtr &operator=(AnyPtr const&ptr) noexcept {
			if (this != &ptr) {
				if (ptr) {
					m_wrapPtr = ptr.m_wrapPtr->copyTo(m_wrapData);
				} else {
					m_wrapPtr = nullptr;
				}
			}
			return *this;
		}
		constexpr operator bool() const noexcept {
			return m_wrapPtr && *m_wrapPtr;
		}
		template<typename T>
		[[nodiscard]]
		constexpr T *get() const noexcept {
#ifdef OX_BARE_METAL
			auto const out = static_cast<Wrap<T>*>(m_wrapPtr);
#else
			auto const out = dynamic_cast<Wrap<T>*>(m_wrapPtr);
#endif
			return out->data;
		}
};

}